from typing import Union

def sort(landmarks: list, reverse: bool) -> list:
    """ O(m+n) integer sorting algorithm developed with the intent of providing better performance than widely-used sorts 
    in terms of both memory and time for integer sorting. Currently runs in O(m_n) time and O(max(m, n)) space.
    
    Parameters
    ----------
    landmarks: List[int]
        List of integers to be sorted
    
    reverse: bool
        True if you want to sort integers in reverses

    
    Returns
    -------
    Pseudo in-place landmarks array. Each value is actually replaced with another integer.
    """
    if len(landmarks) < 1:
        return landmarks
    least, most = find_extremes(landmarks)

    wasteland: list = [0] * (most - least + 1)
    index : int = 0
    for value in landmarks:
        index = value - least
        wasteland[index] += 1

    index = 0


    for x in range(len(wasteland)):
        item: int = wasteland[x]
        while item > 0:
            item -= 1
            if reverse:   
                landmarks[(len(landmarks) - 1) - index] = least + x
            else:
                landmarks[index] = least + x   
            index += 1
        
    



#######Helper Functions#######
def find_extremes(landmarks: list) -> Union[int, int]:
    """ Helper function to find the maximum and minimum elements in a list in O(n) time

    Parameters
    ----------
    landmarks: List[int]
        List of integers to be searched
    

    Returns
    -------
    Two integers: the minimum value and the maximum value in that order
    """
    if landmarks == []:
        raise ValueError("Could not find maxima of the list")

    most, least = landmarks[0], landmarks[0]

    for val in landmarks:
        if val > most:
            most = val
        if val < least:
            least = val
    return least, most



import pytest
from wasteland import wasteland_sort as ws

def test_wasteland_already_sorted():
    landmarks = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    ws.sort(landmarks, False)
    assert landmarks == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

def test_wasteland_out_of_order():
    landmarks = [10, 9, 7, 4, 6, 2, 1, 5, 8, 3]
    ws.sort(landmarks, False)
    assert [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] == landmarks


###
def test_wasteland_reverse_already_sorted():
    landmarks = [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]
    ws.sort(landmarks, True)
    assert landmarks == [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]

def test_wasteland_reverse_out_of_order():
    landmarks = [10, 9, 7, 4, 6, 2, 1, 5, 8, 3]
    ws.sort(landmarks, True)
    assert [10, 9, 8, 7, 6, 5, 4, 3, 2, 1] == landmarks


###
def test_wasteland_large_wastes():
    landmarks = [1,1,1,20,1,1,20,1,1,1,100,1,1,1,1,1,1,1,]
    ws.sort(landmarks, False)
    assert [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,20,20,100] == landmarks

def test_wasteland_large_wastes_reverse():
    landmarks = [1,1,1,20,1,1,20,1,1,1,100,1,1,1,1,1,1,1,]
    ws.sort(landmarks, True)
    assert [100,20,20,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1] == landmarks


###
def test_wasteland_empty():
    assert [] == ws.sort([], False)

def test_wasteland_reverse_empty():
    assert [] == ws.sort([], True)




############Helper functions##################
def test_find_extremes_empty():
    with pytest.raises(ValueError):
        ws.find_extremes([])

def test_find_extremes_nonempty():
    with pytest.raises(ValueError):
        ws.find_extremes([])

def test_find_extremes_one_value():
    assert 1, 1 == ws.find_extremes([1])

def test_find_extremes():
    input = [10, 9, 7, 4, 6, 2, 1, 5, 8, 3]
    assert 1, 10 == ws.find_extremes(input)
    


import setuptools
import pathlib


HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setuptools.setup(
    name="wasteland_sort",
    version="1.1.0",
    author="Markian Hromiak",
    author_email="markian.hromiak@yahoo.com",
    description="Fully tested and documented version of the Wasteland sorting algorithm",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/MHromiak/wasteland_sort.git",
    packages=setuptools.find_packages(exclude=("test",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8.2',    
)
